use core::time::Duration;
use std::collections::HashMap;

use log::{debug, error, info};
use tokio::time::delay_for;

use crate::connectors::{Client, ExchangeDetails};
use crate::currency::{Symbol, SymbolPair};
use crate::events::Event;
use crate::managers::{ExchangeManager, OrderManager, PositionManager, PriceManager};
use crate::strategy::PositionStrategy;
use crate::ticker::Ticker;
use crate::BoxError;

pub struct BfxBot {
    ticker: Ticker,
    quote: Symbol,
    trading_symbols: Vec<Symbol>,
    exchange_managers: Vec<ExchangeManager>,
}

impl BfxBot {
    pub fn new(
        exchanges: Vec<ExchangeDetails>,
        trading_symbols: Vec<Symbol>,
        quote: Symbol,
        tick_duration: Duration,
    ) -> Self {
        let pairs: Vec<_> = trading_symbols
            .iter()
            .map(|x| SymbolPair::new(quote.clone(), x.clone()))
            .collect();

        let exchange_managers = exchanges
            .iter()
            .map(|x| ExchangeManager::new(x, &pairs))
            .collect();

        BfxBot {
            ticker: Ticker::new(tick_duration),
            quote,
            trading_symbols,
            exchange_managers,
        }
    }

    pub async fn start_loop(&mut self) -> Result<(), BoxError> {
        self.update_exchanges().await?;

        loop {
            info!("Current tick: {}", self.ticker.current_tick());

            if let Err(e) = self.update().await {
                error!("Error in main bot loop: {}", e);
            }
        }
    }

    async fn update_exchanges(&mut self) -> Result<(), BoxError> {
        for e in &mut self.exchange_managers {
            if let Err(err) = e.update_managers(self.ticker.current_tick()).await {
                error!("Error while updating managers: {}", err);
            }
        }

        Ok(())
    }

    async fn update(&mut self) -> Result<(), BoxError> {
        delay_for(self.ticker.duration()).await;
        self.ticker.inc();

        self.update_exchanges().await?;

        Ok(())
    }
}
